from abc import ABC, abstractmethod, ABCMeta, abstractproperty


class Singleton(ABC):
    _instance = None

    def __new__(cls, *args, **kwargs):
        if cls._instance is None:
            cls._instance = super(Singleton, cls).__new__(cls)
        return cls._instance



from bottlenest.common import Module
from bottlenest.transports.cli import CommandFactory
from bottlenest.corecommands.NewCoreCommand import NewCoreCommand
from bottlenest.corecommands.BuildCoreCommand import BuildCoreCommand


@Module(
    providers=[NewCoreCommand, BuildCoreCommand],
)
class CoreCliModule:
    pass


def main():
    CommandFactory.run(CoreCliModule)


if __name__ == "__main__":
    main()



from bottlenest.metaClasses.NestProvider import NestProvider
from .NestRoute import NestRoute

# TODO: add support for route prefix


class NestController(NestProvider):
    __name__ = 'NestController'

    def eventName(self):
        return NestRoute.__name__



class NestCommandArgument:
    __name__ = "NestCommandArgument"

    def __init__(
        self,
        argumentName,
        callback,
        optional,
        *args,
        **kwargs,
    ):
        self.argumentName = argumentName
        self.callback = callback
        self.optional = optional
        self.args = args
        self.kwargs = kwargs



class HttpError(Exception):
    def __init__(self, messages=[], statusCode=500):
        message = messages if isinstance(messages, str) else messages[0]
        super(HttpError, self).__init__(message)
        self.messages = messages
        self.statusCode = statusCode

    def toDict(self):
        # check if messages is a list
        if isinstance(self.messages, list):
            return {'messages': self.messages}
        return {'messages': [str(self.messages)]}

    def __str__(self):
        return self.messages

    def __repr__(self):
        return self.messages



import os
import shutil


class FileUtils:
    @staticmethod
    def absPath(path):
        # this function should be the abs path relative to the project root
        basePath = os.path.dirname(os.path.dirname(os.path.abspath(__file__)))
        return os.path.join(basePath, path)

    @staticmethod
    def copyDir(src, dst, symlinks=False, ignore=None):
        # copy dir
        # do not use for
        shutil.copytree(src, dst, symlinks, ignore)



import eventlet
import socketio
from ...http.HttpTransport import HttpTransport


class WebsocketsFactory:
    __gateways__: dict[str, dict] = {}
    __appContext__ = None

    @staticmethod
    def setAppContext(appContext) -> None:
        print("---x-x--x--x-------------x-x--x--------------> setAppContext")
        if WebsocketsFactory.__appContext__ is None:
            WebsocketsFactory.__appContext__ = appContext

    @staticmethod
    def registerGateway(provider, providerContext) -> None:
        # add gateway
        sio = socketio.Server()
        providerContext.set('sio', sio)
        WebsocketsFactory.__gateways__[provider.getName()] = {
            'provider': provider,
            'providerContext': providerContext,
            'sio': sio,
            # 'app': httpTransport.app,
            # 'port': provider.port,
        }

    @staticmethod
    def setupGateway(gatewayDict):
        sio = gatewayDict['sio']
        provider = gatewayDict['provider']
        providerContext = gatewayDict['providerContext']
        port = provider.port
        httpTransport = HttpTransport(port=provider.port)
        httpTransport.setupTransport(
            moduleContext=providerContext,
            appContext=WebsocketsFactory.__appContext__,
        )
        # app = socketio.WSGIApp(sio)

        def _onConnect(sid, environ, auth):
            print(f"[NestWebsocketsFactory] onConnect {sid}")
        sio.on('connect')(_onConnect)

        app = socketio.WSGIApp(sio, httpTransport.app)
        eventlet.wsgi.server(eventlet.listen(('0.0.0.0', port)), app)
        print(f"[listen] Starting websockets server on port {port}")
        # pool.spawn(eventlet.wsgi.server,
        #           eventlet.listen(("", port)), app)
        # run pool.spawn with Flask (app) and eventlet.listen(("", port))
        # return pool.spawn(eventlet.wsgi.server,
        #                   eventlet.listen(("0.0.0.0", port)), app)

    @staticmethod
    def listen(pool) -> None:
        print("Listen WEBSOCKETSSSSSS now!")

        def startServer(gateways):
            for gateway in gateways:
                pool.spawn(WebsocketsFactory.setupGateway, gateway)

        return startServer(WebsocketsFactory.__gateways__.values())
        # asyncio.run(asyncio.
        # eventlet.wsgi.server(eventlet.listen(('', port)), app)



from flask import request
from functools import wraps


class NestRoute:
    __name__ = 'NestRoute'

    def __init__(self, path, method, callback):
        print(f"route defined {method} {path}")
        self.callback = callback
        self.path = self._nestjsToFlaskPath(path)
        self.method = method

    def setupEvent(self, cls, moduleContext):
        print(f"NestRoute setupEvent {self.path}")
        print(moduleContext)
        transport = moduleContext.get('transport')
        app = transport.app
        # app = moduleContext.get('app')
        app.add_url_rule(
            self.path,
            methods=[self.method],
            view_func=self._callbackWrapper(cls),
        )

    def _callbackWrapper(self, cls):
        @wraps(self.callback)
        def wrapped(*args, **kwargs):
            return self.callback(cls, NestRequest(request))
            # return self.callback(context)
        return wrapped

    # Flask routing is different than NestJS routing
    def _nestjsToFlaskPath(self, path: str) -> str:
        result = ''
        parts = path.split('/')
        for part in parts:
            if part.startswith(':'):
                result += '/<' + part[1:] + '>'
            else:
                result += '/' + part
        return result[1:]

##################################################################


class NestRequest:
    __name__ = 'NestRequest'

    def __init__(self, request):
        self.request = request
        self.params = NestRequestParams(self.request)
        self.query = {}
        self.body = request.get_json(silent=True, force=True)
        self.headers = {}

        if self.body is None:
            self.body = {}


class NestRequestParams(object):
    __name__ = 'NestRequestParams'

    def __init__(self, request):
        super(NestRequestParams, self).__init__()
        self.request = request

    def __getattribute__(self, __name: str):
        if __name == 'request':
            return super(NestRequestParams, self).__getattribute__(__name)
        else:
            return self.request.view_args[__name]
            # return self.request.args.get(__name, type=str)



from bottlenest.common.Injectable import *

# auto import from http
from bottlenest.transports.http.decorators import *
from bottlenest.transports.http.errors import *
from bottlenest.transports.http import *



from bottlenest.metaClasses.NestProvider import NestProvider
from .NestSubscribeMessage import NestSubscribeMessage
from ..factories.WebsocketsFactory import WebsocketsFactory

servers = []


class NestWebSocketGateway(NestProvider):
    __name__ = 'NestWebSocketGateway'

    def __init__(self, cls, port=4001, namespace=None):
        self.cls = cls
        self.providerName = cls.__name__
        self.port = port
        self.namespace = namespace

    def getName(self):
        return self.cls.__name__

    def eventName(self):
        return NestSubscribeMessage.__name__

    # called from whithin the module's setup
    def setupProvider(self, module, context):
        print(f"NestSocketGateway setupProvider {self.providerName}")
        self.module = module
        self.context = context
        # get sio from WebsocketsTransport
        WebsocketsFactory.registerGateway(self, context)
        self._setupProvider(module, context)

    def listen(self):
        print(f"NestSocketGateway listen {self.providerName}")



from .factories.WebsocketsFactory import WebsocketsFactory


class WebsocketsTransport:
    # called by user, passing any options
    def __init__(self):
        pass

    def setupTransport(self, appContext, moduleContext):
        print("================== setup WebsocketsTransport")
        # self.appContext = appContext
        # self.module = appContext.module
        # self.logger = appContext.logger
        WebsocketsFactory.setAppContext(appContext)
        pass

    def listen(self, callback):
        WebsocketsFactory.listen()
        callback()

    # def listen(self, callback):
    #     # TODO: review this port to allow multiple ports
    #     port = self.appContext.get('port')
    #     app = self.appContext.get('app')
    #     eventlet.wsgi.server(eventlet.listen(('', port)), app)
    #     callback()



from bottlenest.core.NestApplicationContext import NestApplicationContext
import os


class NestFactory:
    @staticmethod
    def createMicroservice(module, transport=None):
        instance = NestFactory.createApplicationContext(
            module=module,
            transport=transport,
        )
        return instance

    @staticmethod
    def createApplicationContext(module, transport):
        return NestApplicationContext(
            module=module,
            transport=transport,
        )



from bottlenest.core.NestLogger import *
from bottlenest.core.NestFactory import *
from bottlenest.core.NestApplicationContext import *
from bottlenest.core.NestModuleContext import *
from bottlenest.common import *



import py_compile
from bottlenest.transports.cli import Command, CommandArgument


@Command(
    name="build",
    description="Generates a build of the project",
)
class BuildCoreCommand:
    __name__ = 'BuildCoreCommand'

    def __init__(self, context):
        self.context = context

    def run(self, context, args):
        print(f"{self.__name__}.run:")
        print(f"{args.path} --> {args.output}")
        py_compile.compile(args.path, args.output)

    @CommandArgument(
        name="--path",
        dest="path",
        help="The path to the file to compile",
        default="examples/clitool/main.py",
        optional=True,
    )
    def pathArg(self, value):
        return value

    @CommandArgument(
        name="--output",
        dest="output",
        help="The output file",
        default="compiled.pyc",
        optional=True,
    )
    def outputArg(self, value):
        return value
